#include <v_lib.h>
#include "eval.h"
#include "cspace.h"

#define	ARG_MATRIX	0
#define	ARG_VECTOR	1

void
matZeroRows(matrix_p m,u32_t row_start,u32_t row_end)
{
	u32_t r, row_len;
	V_ASSERT(m);
	if(row_start > row_end || row_start > m->rows)
		return;
	if(row_end > m->rows)
		row_end = m->rows;
	row_len = m->cols * sizeof(vec_t);
	for(r = row_start;r < row_end;r++)
		memZero(m->cell[r],row_len);
}

void
matZeroCols(matrix_p m,u32_t col_start,u32_t col_end)
{
	V_ASSERT(m);
	if(m->rows == m->cols)
	{
		matTransposeIP(m);
		matZeroRows(m,col_start,col_end);
		matTransposeIP(m);
	}
	else
	{
		matrix_p transposed = matAlloc(m->cols,m->rows);
		matTranspose(transposed,m);
		matZeroRows(transposed,col_start,col_end);
		matTranspose(m,transposed);
		memFree(transposed);
	}
}

void
vecRunningSum(vector_p dest,vector_p src)
{
	u32_t i;
    vec_p d = dest->elements, s = src->elements;
	V_ASSERT(src->size == dest->size);
    d[0] = s[0];
    for(i = 1;i < dest->size;i++)
    	d[i] = s[i] + d[i-1];
}

void
vecLinSpaced(vector_p dest,vec_t min,vec_t max)
{
	u32_t i, n_divs;
    vec_t delta;
    n_divs = dest->size;
    delta = (max - min) / (n_divs - 1);
    dest->elements[0] = min;
    for(i = 1;i < n_divs;i++)
    	dest->elements[i] = dest->elements[i-1] + delta;
}

void
matScaleAdd(matrix_p dest,matrix_p src,vec_t scale,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = offset + src->cell[r][c] * scale;
}

/* *********************************************************** */

vp_stack_t	cached_args = TYPED_STACK_EMPTY_INITIALIZER;

V_LOCAL arg_p
evalSearchForName(char *id)
{
	u32_t i, count = vpsCount(&cached_args);
	arg_p *args = (arg_p*)vpsGetStorage(&cached_args);
	for(i=0;i<count;i++)
		if(args[i] && STR_EQ(args[i]->id,id))
			return args[i];
	return NULL;
}

arg_p
evalAllocArg(char *name,int type)
{
	arg_p arg;
	if(evalSearchForName(name))
	{
		conlPrintf(LOG_ERROR,"evalAllocArg(%s): duplicate definition\n",name);
		return NULL;
	}
	arg = (arg_p)memAllocZ(sizeof(arg_t));
	V_ASSERT(arg);
	stateDefault(&(arg->state));
    strCopy(arg->id,name,MAX_ID);
    arg->type = type;
	arg->evaluated = false;
	vpsPush(&cached_args,arg);
	return arg;
}

void
evalFreeArg(arg_p arg)
{
	u32_t i, count = vpsCount(&cached_args);
	arg_p *args = (arg_p*)vpsGetStorage(&cached_args);
	for(i=0;i<count;i++)
		if(args[i] == arg)
			args[i] = NULL;
	if(arg->op.data)
		memFree(arg->op.data);
	memFree(arg);
}

void
evalFreeAllArgs(void)
{
	u32_t i, count = vpsCount(&cached_args);
	arg_p *args = (arg_p*)vpsGetStorage(&cached_args);
	for(i=0;i<count;i++)
	{
		if(args[i])
		{
			if(args[i]->op.data)
				memFree(args[i]->op.data);
			memFree(args[i]);
		}
	}
	vpsClear(&cached_args);
}

matrix_p
evalMatrix(char *name,vhandle_t script)
{
	arg_p arg;
	arg = evalSearchForName(name);
	if(arg)
	{
		if(arg->type != ARG_MATRIX)
		{
			conlPrintf(LOG_ERROR,"evalMatrix(%s): object of another type "
				"already exists.\n",name);
			return NULL;
		}
		if(!arg->evaluated)
		{
			conlPrintf(LOG_ERROR,"evalMatrix(%s): circular reference.\n",
				name);
			return NULL;
		}
		return arg->op.matrix;
	}
    arg = evalAllocArg(name,ARG_MATRIX);
    if(!arg)
    	return NULL;
	tokPush(script);
    arg->evaluated = scriptEvalMatrix(arg,script);
	tokPop(script);
    if(!arg->evaluated)
    {
    	evalFreeArg(arg);
    	return NULL;
    }
    return arg->op.matrix;
}

vector_p
evalVector(char *name,vhandle_t script)
{
	arg_p arg;
	arg = evalSearchForName(name);
	if(arg)
	{
		if(arg->type != ARG_VECTOR)
		{
			conlPrintf(LOG_ERROR,"evalVector(%s): object of another type "
				"already exists.\n",name);
			return NULL;
		}
		if(!arg->evaluated)
		{
			conlPrintf(LOG_ERROR,"evalVector(%s): circular reference.\n",
				name);
			return NULL;
		}
		return arg->op.vector;
	}
    arg = evalAllocArg(name,ARG_VECTOR);
    if(!arg)
    	return NULL;
    arg->evaluated = scriptEvalVector(arg,script);
    if(!arg->evaluated)
    {
    	evalFreeArg(arg);
    	return NULL;
    }
    return arg->op.vector;
}

vboolean
evalColorChannel(vhandle_t script,matrix_p channels[],
	u32_t index,
	int source_space,int assumed_space)
{
	char name[MAX_ID] = "";
	if(source_space != assumed_space)
	{
		conlPuts(LOG_ERROR,
        	"evalColorChannel(): color space mismatch\n");
		return false;
	}
	if(!tokNext(script))
	{
		conlPuts(LOG_ERROR,
        	"evalColorChannel(): name expected, got EOF\n");
		return false;
	}
	strCopy(name,tokLast(script),MAX_ID);
	channels[index] = evalMatrix(script,name);
	return (channels[index]) ? true : false;
}

vboolean
evalOutput(char *file_name,vhandle_t script)
{
	framebuf_p fb = NULL;
	u32_t source_space = CSPACE_RGB;
	matrix_p channels[3] = {NULL,NULL,NULL};
	matrix_p alpha = NULL;
	while(tokNext(script))
    {
    	char *last = dbs(tokLast(script));
    	if(STR_EQ(last,"}"))
        	break;
		else if(STR_EQ(last,"source_space"))
        {
        	char *space = tokNext(script);
            V_ASSERT(space);
        	if(STR_EQ(space,"RGB"))
            	source_space = CSPACE_RGB;
			else if(STR_EQ(space,"HSV"))
            	source_space = CSPACE_HSV;
			else if(STR_EQ(space,"CIE"))
            	source_space = CSPACE_CIE;
			else
            {
            	conlPrintf(LOG_ERROR,
                	"evalOutput(): unknown color space '%s'\n",space);
                return false;
            }
        }
        else if(STR_EQ(last,"red"))
        {
        	if(!evalColorChannel(script,channels,0,source_space,CSPACE_RGB))
        		goto eval_error;
        }
        else if(STR_EQ(last,"green"))
        {
        	if(!evalColorChannel(script,channels,1,source_space,CSPACE_RGB))
        		goto eval_error;
        }
        else if(STR_EQ(last,"blue"))
        {
        	if(!evalColorChannel(script,channels,2,source_space,CSPACE_RGB))
        		goto eval_error;
        }
        else if(STR_EQ(last,"alpha"))
        {
        	if(!evalColorChannel(script,&alpha,0,CSPACE_RGB,CSPACE_RGB))
        		goto eval_error;
        }
        else if(STR_EQ(last,"hue"))
        {
        	if(!evalColorChannel(script,channels,0,source_space,CSPACE_HSV))
        		goto eval_error;
        }
        else if(STR_EQ(last,"saturation"))
        {
        	if(!evalColorChannel(script,channels,1,source_space,CSPACE_HSV))
        		goto eval_error;
        }
        else if(STR_EQ(last,"value"))
        {
        	if(!evalColorChannel(script,channels,2,source_space,CSPACE_HSV))
        		goto eval_error;
        }
        else if(STR_EQ(last,"cieX"))
        {
        	if(!evalColorChannel(script,channels,0,source_space,CSPACE_CIE))
        		goto eval_error;
        }
        else if(STR_EQ(last,"cieY"))
        {
        	if(!evalColorChannel(script,channels,1,source_space,CSPACE_CIE))
        		goto eval_error;
        }
        else if(STR_EQ(last,"cieZ"))
        {
        	if(!evalColorChannel(script,channels,2,source_space,CSPACE_CIE))
        		goto eval_error;
        }
        else
        {
    		conlPrintf(LOG_ERROR,
    			"evalOutput(%s): unknown channel '%s'.\n",
				file_name,last);
        	return false;
        }
    }
    if(channels[0] && channels[1] && channels[2])
    {
    	if(channels[0]->rows != channels[1]->rows ||
    		channels[0]->rows != channels[2]->rows ||
    		channels[0]->cols != channels[1]->cols ||
    		channels[0]->cols != channels[2]->cols)
    	{
    		conlPrintf(LOG_ERROR,
    			"evalOutput(%s): channels of different size.\n",
				file_name);
    		return false;
    	}
    	fb = fbAlloc(channels[0]->cols,channels[0]->rows);
    	V_ASSERT(fb);
    	ctransStore3Matrices(fb,source_space,
    		channels[0],channels[1],channels[2]);
	    if(alpha)
	    {
	    	if(channels[0]->rows == alpha->rows &&
	    		channels[0]->cols == alpha->cols)
	    		ctransStore1Matrix(fb,alpha,false,false,false,true);
	    	else
    			conlPrintf(LOG_WARNING,
    				"evalOutput(%s): alpha of different size: ignored.\n",
					file_name);
	    }
	    fbSave(fb,file_name);
	    memFree(fb);
	    return true;
    }
	conlPrintf(LOG_ERROR,"evalOutput(%s): not enought channels.\n",
		file_name);
	return false;
eval_error:;
	conlPrintf(LOG_ERROR,"evalOutput(%s): unable to eval channel.\n",
		file_name);
	return false;
}

vboolean
evalAll(vhandle_t script)
{
	while(tokSkipToEx(script,"output",STR_COMP))
	{
		char name[MAX_ID];
		if(!tokNext(script))
		{
			conlPuts(LOG_ERROR,
        		"evalAll(): output file name expected, got EOF\n");
			return false;
		}
		strCopy(name,tokLast(script),MAX_ID);
		if(!evalOutput(script,name))
			return false;
	}
	return true;
}

/* *********************************************************** */

vboolean
scriptSearchPair(vhandle_t script,char *key,char *name)
{
	tokRestart(script);
	while(tokSkipToEx(script,key,STR_COMP))
		if(tokNextIsEx(script,name,STR_COMP))
			return true;
	return false;
}

#define SYNTAX_ERROR(script,caller,error_desc) \
	conlPrintf(LOG_ERROR,"%s(): line %s: 's': %s.\n",caller, \
		tokLastLine(script),dbs(tokLast(script)),error_desc)
	

/* *********************************************************** */

/*
 * Matrix operations
 */

vboolean matrix_input_op(arg_p arg,vhandle_t script)
{
	char img_name[256];
	strCopy(img_name,dbs(tokLast(script)),256);
	arg->op.matrix = inputGetMatrix(&arg->state,img_name,tokNext(script));
	return (arg->op.matrix) ? true : false;
}

vboolean matrix_save_op(arg_p arg,vhandle_t script)
{
	vhandle_t outfile;
	char file_name[_MAX_PATH];
	strCopy(file_name,dbs(tokLast(script)),_MAX_PATH);
	outfile = fileOpenWrite(file_name);
	if(outfile == LIBV_INVALID_HANDLE)
		return false;
	
	// todo
	fileClose(outfile);
	return true;
}

/* *********************************************************** */

/*
 * Vector operations
 */

/* *********************************************************** */

typedef vboolean (*matrix_op)(arg_p,vhandle_t);

typedef vboolean (*vector_op)(arg_p,vhandle_t);

typedef struct
{
	char 		*op_name;
	matrix_op	routine;
	char		*description;
}	matrix_op_field;

typedef struct
{
	char 		*op_name;
	vector_op	routine;
	char		*description;
}	vector_op_field;

matrix_op_field matrix_op_fields[] =
{
	{"input",matrix_input_op,"reads matrix from image file; requires image name and channel name"},
	{"save",matrix_save_op,"writes matrix to text file; requires file name"},
	{NULL,NULL,NULL}
};

vector_op_field vector_op_fields[]=
{
	{NULL,NULL,NULL}
};

vboolean
evalFindMatrixOp(char *op,matrix_op *proc_ret)
{
	matrix_op_field *ptr;
	for(ptr = matrix_op_fields;ptr->op_name;ptr++)
		if(STR_EQ(ptr->op_name,op))
		{
			*proc_ret = ptr->routine;
			return true;
		}
	return false;
}

vboolean
evalFindVectorOp(char *op,vector_op *proc_ret)
{
	vector_op_field *ptr;
	for(ptr = vector_op_fields;ptr->op_name;ptr++)
		if(STR_EQ(ptr->op_name,op))
		{
			*proc_ret = ptr->routine;
			return true;
		}
	return false;
}

void
evalHelpMatrixOps(void)
{
	matrix_op_field *ptr;
	conPuts("Available matrix operations:\n");
	for(ptr = matrix_op_fields;ptr->op_name;ptr++)
		conPrintf("\t'%s' - %s\n",ptr->op_name,dbs(ptr->description));
}

void
evalHelpVectorOps(void)
{
	vector_op_field *ptr;
	conPuts("Available vector operations:\n");
	for(ptr = vector_op_fields;ptr->op_name;ptr++)
		conPrintf("\t'%s' - %s\n",ptr->op_name,dbs(ptr->description));
}

void evalHelp()
{
	evalHelpMatrixOps();
	evalHelpVectorOps();
}

/* *********************************************************** */

vboolean
scriptEvalMatrix(arg_p arg,vhandle_t script)
{
	matrix_op op_routine;
	V_ASSERT(arg->type == ARG_MATRIX);
	if(!scriptSearchPair(script,"matrix",arg->id))
	{
		conlPrintf(LOG_ERROR,"scriptEvalMatrix(%s): undefined.\n",
			arg->id);
		return false;
	}
	if(!tokNextIs(script,"{"))
	{
		SYNTAX_ERROR(script,"scriptEvalMatrix","expected '{'");
		return false;
	}
	while(tokNext(script))
	{
		char last[256];
		strCopy(last,tokLast(script),256);
		if(STR_EQ(last,"}"))
		{
			break;
		} else if(STR_EQ(last,"state"))
		{
			if(!stateChange(&(arg->state),script))
				break;
		} else if(evalFindMatrixOp(last,&op_routine))
		{
			if(!(*op_routine)(arg,script))
				break;
		} else
		{
			/* no commands: assume input */
			arg->op.matrix = evalMatrix(script,last);
		}
	}
	return (arg->op.matrix) ? true : false;
}

/* *********************************************************** */

vboolean
scriptEvalVector(arg_p arg,vhandle_t script)
{
	vector_op op_routine;
	V_ASSERT(arg->type == ARG_VECTOR);
	if(!scriptSearchPair(script,"vector",arg->id))
	{
		conlPrintf(LOG_ERROR,"scriptEvalVector(%s): undefined.\n",
			arg->id);
		return false;
	}
	if(!tokNextIs(script,"{"))
	{
		SYNTAX_ERROR(script,"scriptEvalVector","expected '{'");
		return false;
	}
	while(tokNext(script))
	{
		char last[256];
		strCopy(last,tokLast(script),256);
		if(STR_EQ(last,"}"))
		{
			break;
		} else if(STR_EQ(last,"state"))
		{
			if(!stateChange(&(arg->state),script))
				break;
		} else if(evalFindVectorOp(last,&op_routine))
		{
			if(!(*op_routine)(arg,script))
				break;
		} else
		{
			/* no commands: assume input */
			arg->op.vector = evalVector(script,last);
		}
	}
	return (arg->op.vector) ? true : false;
}
